class SBQueenSaver expands Mutator config(SBQueenSaver);


enum ELogType {
	LOG_None,
	LOG_Error,
	LOG_Warning,
	LOG_Info,
	LOG_Verbose,
	LOG_Debug,
	LOG_All,
};

struct Spawner {
	var string ClassName, PropertyName, PropertyValue;
};

struct DontTreatThisActor {
	var string MapName, ActorName;
};


var config Spawner Spawners[16];
var config DontTreatThisActor DontTreatTheseActors[32];

var config ELogType LogLevel;
var config bool	bReplaceQueenSubClassesAlso; // Replace sub classes of Queen also (or class'Queen' only)
var config bool	bActive;
var config bool	bUseLogTimeStamp; // Add a time stamp in front of every log message?
var config bool	bUseStaticQueenDestsCount; // If true, function QueenDestsHave() will cache the result of first QueenDests search.

/************************** internal variables **************************/
const CMutatorVersion = 1;
const CMutatorName = "SBQueenSaver";
const CNewQueenClassName = "Class'SBQueenSaver-V1.SBQueen'";
const CNewQueenFullName = "SBQueenSaver-V1.SBQueen";
var bool	bQueenDestsHave; // is set to TRUE by HaveQueenDest() if at least one QueenDest exists in map
var bool	bQueenDestsHaveIsSet; // is set to TRUE by function HaveQueensDest() on first run -> bQueenDestsHave is valid
var byte	DontTreatTheseActorsCount;
var string	MapFileName;



function bool CheckReplacement(Actor Other, out byte bSuperRelevant) {
/******************************************************************************
The result of CheckReplacement is the result for IsRelevent().
Therefore return TRUE to keep *Other*, return FALSE to have it deleted.
Set *bSuperRelevant=byte(True)* to skip further checks in GameInfo.IsRelevant().
******************************************************************************/
local SBQueen NewQueen;

	if (Queen(Other) != None && (Other.Class == Class'Queen' || bReplaceQueenSubClassesAlso) && Other.Class != Class'SBQueen')
	{
		if ( ! MayTreatThisActor(Other.Name))
			return true;

		NewQueen = SBQueen(SpawnAndPasteActor(Other, CNewQueenFullName));
		if (NewQueen != None)
		{
			class'SharedCode'.static.CopyProperties_Queen(Queen(Other), NewQueen, true);
			if (QueenDestsHave( ! bUseStaticQueenDestsCount))
				NewQueen.QueenDestHave = QDH_Yes;
			else
				NewQueen.QueenDestHave = QDH_No;

			if (Queen(Other).MenuName == "")
				NewQueen.MenuName = NewQueen.Default.MenuName;
			if (Other.DrawScale != class'Queen'.default.DrawScale)
			{
				class'SharedCode'.static.CopyProperties_Queen(Queen(Other), NewQueen, true);
				logger(LOG_Debug, "CheckReplacement", NewQueen $ ".Drawsize=" $ NewQueen.Drawsize @ "is not the default DrawSize, changing Collision Cylinder");
				CollisionCylinderAdjust(NewQueen, false);
				logger(LOG_Debug, "CheckReplacement", "Collision Cylinder changed to" @ NewQueen $ ".CollisionRadius=" $ NewQueen.CollisionRadius $ ", CollisionHeight=" $ NewQueen.CollisionHeight);
			}
			else
				logger(LOG_Debug, "CheckReplacement", NewQueen @ " has default Drawsize, therefore collision cylinder not changed");
			bSuperRelevant = 0;
			logger(LOG_Info, "CheckReplacement", "replaced" @ Other @ "with" @ NewQueen);
			return false;
		}
		else
			logger(LOG_Error, "CheckReplacement", "spawning of" @ CNewQueenClassName @ "failed");

	}
	return true;
}



function bool CollisionCylinderAdjust(Actor A, optional bool bForce) {
/******************************************************************************
Adjusts the CollisionCylinder if *DrawScale* is changed and *CollisionRadius*
and *CollisionHeight* is unchanged or *bForce*==true.

Returns true if *A* was modified.
******************************************************************************/
local float NewRadius;
local float NewHeight;

	if ( ! bForce)
	{
		if (A.DrawScale == A.Default.DrawScale) return false;
		// has mapper already adjusted collisioncylinder?
		if (abs(A.CollisionRadius - A.Default.CollisionRadius) > 0.1) return false;
		if (abs(A.CollisionHeight - A.Default.CollisionHeight) > 0.1) return false;
	}

	return A.SetCollisionSize(A.Default.CollisionRadius * A.DrawScale, A.Default.CollisionHeight * A.DrawScale);
}



function byte DontTreatTheseActorsInit() {
/******************************************************************************
Removes all gaps and invalid entries from the array *DontTreatTheseActors*
and returns the number of valid entries.
******************************************************************************/
local byte i, j, result;
local bool RemoveLine;

	// wipe out invalid entries and detect if complete map should be ignored:
	for (i = 0; i < ArrayCount(DontTreatTheseActors); i++)
	{
		if (DontTreatTheseActors[i].MapName == "")
		{
			if (DontTreatTheseActors[i].ActorName != "")
				logger(LOG_Warning, "DontTreatTheseActorsInit", "empty 'MapName' but non empty 'ActorName' in row" @ i @ ", entry ignored");
			Goto ClearItem;
		}
		else if (DontTreatTheseActors[i].ActorName == "") // don't replace anything in this map
		{
			if (MapFileName ~= DontTreatTheseActors[i].MapName)
			{
				logger(LOG_Info, "DontTreatTheseActorsInit", "empty 'ActorName' in row" @ i @ "and 'MapName' matches current map '" $ MapFileName $ "', deactivating mutator");
				bActive = false;
				Return 0;
			}
			else // remove item
			{
				logger(LOG_Debug, "DontTreatTheseActorsInit", "empty 'ActorName' in row" @ i @ "but given MapName='" $ DontTreatTheseActors[i].MapName $ "' -> Mutator will be deactivated for that map");
				Goto ClearItem;
			}
		}
		else if (MapFileName ~= DontTreatTheseActors[i].MapName) // entry ok, MapName is current map and ActorName is given
			Goto NextLoop;
		else // MapName does not match -> ignore this entry
		{
			logger(LOG_Debug, "DontTreatTheseActorsInit", "MapName='" $ DontTreatTheseActors[i].MapName $ "' in row" @ i @ "does not match current map name '" $ MapFileName $ "', removing row");
			Goto ClearItem;
		}

ClearItem:
		DontTreatTheseActors[i].MapName = "";
		DontTreatTheseActors[i].ActorName = "";
NextLoop:
	}

	// compact the array:
	for (i = 0; i < ArrayCount(DontTreatTheseActors) - 1; i++)
		if (DontTreatTheseActors[i].MapName == "")
		{
			if (DontTreatTheseActorsHaveNonEmptyRow(i + 1, j)) // swap the rows
			{
				DontTreatTheseActors[i].MapName = DontTreatTheseActors[j].MapName;
				DontTreatTheseActors[j].MapName = "";
				DontTreatTheseActors[i].ActorName = DontTreatTheseActors[j].ActorName;
				DontTreatTheseActors[j].ActorName = "";
				result++;
			}
			else
				break; // nothing to do anymore
		}
		else
			result++;

	if (LogLevel >= LOG_Debug)
		if (result == 0)
			logger(LOG_Debug, "DontTreatTheseActorsInit", "array 'DontTreatTheseActors' has no entries left");
		else
		{
			logger(LOG_Debug, "DontTreatTheseActorsInit", "Compacted array 'DontTreatTheseActors' with" @ result @ "row(s):");
			for (i = 0; i < result; i++)
				logger(LOG_Debug, "DontTreatTheseActorsInit", "DontTreatTheseActors[" $ i $ "].MapName='" $ DontTreatTheseActors[i].MapName $ "', DontTreatTheseActors[" $ i $ "].ActorName='" $ DontTreatTheseActors[i].ActorName $ "'");
		}

	Return result;
}



function bool DontTreatTheseActorsHaveNonEmptyRow(byte StartRow, out byte NonEmptyRow) {

	for (NonEmptyRow = StartRow; NonEmptyRow < ArrayCount(DontTreatTheseActors); NonEmptyRow++)
		if (DontTreatTheseActors[NonEmptyRow].MapName != "")
			return true;
	return false;
}



function String GetSpawnerDebugText(byte index) {

	return "Spawners[" $ index $ "]: ClassName=" $ Spawners[index].ClassName $ ", PropertyName=" $ Spawners[index].PropertyName $ ", PropertyValue=" $ Spawners[index].PropertyValue;
}



function Logger(ELogType NeededLogLevel, string FunctionName, string msg) {
/*******************************************************************************
Encapsulation for SharedCode.Logger().
*******************************************************************************/
	if (LogLevel >= NeededLogLevel)
		class'SharedCode'.static.LogEx(self, bUseLogTimeStamp, FunctionName @ GetEnum(Enum'ELogType', NeededLogLevel), msg);
}



function Bool MayTreatThisActor(coerce string ActorName) {
local byte i;

	if ( ! bActive)
		Goto Forbidden;
	for (i = 0; i < DontTreatTheseActorsCount; i++)
		if (DontTreatTheseActors[i].ActorName ~= ActorName)
			Goto Forbidden;;
	Goto Allowed;

Allowed:
	Logger(LOG_Debug, "MayTreatThisActor", "returns TRUE for" @ ActorName);
	return true;
Forbidden:
	Logger(LOG_Debug, "MayTreatThisActor", "returns FALSE for" @ ActorName);
	return false;
}



event PostBeginPlay() {
local string MutatorInfoStr;
local Actor A;
local String OldPropertyValue;
local ThingFactory TF;
local int i, Count;

	Super.PostBeginPlay();
	MutatorInfoStr = "Mutator" @ CMutatorName @ "Version" @ CMutatorVersion;
	if(bActive)
		Logger(LOG_Info, "PostBeginPlay", MutatorInfoStr @ "loaded and active");
	else
	{
		Logger(LOG_Warning, "PostBeginPlay", MutatorInfoStr @ "disabled via INI settings");
		return;
	}

	// set global variable *MapFileName*:
	MapFileName = string(Level.Outer.Name);

	DontTreatTheseActorsCount = DontTreatTheseActorsInit();


// replace Queen in spawn things:

	// replace in ThingFactory and its sub classes (CreatureFactory, KeypointCreatureFactory, ...)
	logger(LOG_Debug, "PostBeginPlay", "checking for Queens in factories");
	foreach AllActors(class'ThingFactory', TF)
	{
		Logger(LOG_All, "PostBeginPlay", "processing" @ TF $ ", Prototype=" $ TF.ProtoType);
		if (TF.ProtoType != None && MayTreatThisActor(TF))
			if (TF.ProtoType == class'Queen' || (ClassIsChildOf(TF.ProtoType, class'Queen') && bReplaceQueenSubClassesAlso))
			{
				OldPropertyValue = String(TF.ProtoType);
				TF.ProtoType = Class'SBQueen';
				Logger(LOG_Info, "PostBeginPlay", "replaced" @ TF $ ".Prototype=" $ OldPropertyValue @ "with" @ TF.ProtoType);
				Count++;
			}
	}
	logger(LOG_Info, "PostBeginPlay", Count @ "Queen(s) in factories replaced");

	logger(LOG_Debug, "PostBeginPlay", "checking for replacements in 'Spawners':");

	for (i = 0; i < ArrayCount(Spawners); i++)
	{
		Logger(LOG_Debug, "PostBeginPlay", "processing" @ GetSpawnerDebugText(i));
		if (Spawners[i].ClassName != "" && Spawners[i].PropertyName != "" && Spawners[i].PropertyValue != "")
		{
			Count = 0;
			foreach AllActors(class'Actor', A)
				if (string(A.Class.name) ~= Spawners[i].ClassName && A.GetPropertyText(Spawners[i].PropertyName) ~= Spawners[i].PropertyValue)
				{
					OldPropertyValue = A.GetPropertyText(Spawners[i].PropertyName);
					Class'SharedCode'.static.SetPropertyTextEx(A, Spawners[i].PropertyName, CNewQueenClassName);
					Logger(LOG_Info, "PostBeginPlay", "replaced" @ A $ "." $ Spawners[i].PropertyName $ "=" $ OldPropertyValue @ "with" @ A.GetPropertyText(Spawners[i].PropertyName));
					Count++;
				}
				else
					Logger(LOG_All, "PostBeginPlay", A @ "does not match" @ GetSpawnerDebugText(i));
			logger(LOG_Info, "PostBeginPlay", Count @ "Queen(s) in" @ Spawners[i].ClassName @ "replaced");
		}
		else
			Logger(LOG_Debug, "PostBeginPlay", GetSpawnerDebugText(i) @ "has at least one empty entry, skipping");

	}
}



function bool QueenDestsHave(bool force) {
/******************************************************************************
Disabled search via NavigationPointList because later added QueenDests are not
part of it.
If *force*, the last search result is not used.
******************************************************************************/
local QueenDest QD;

	if (bQueenDestsHaveIsSet && ! force)
		return bQueenDestsHave;

	bQueenDestsHave = false;
	bQueenDestsHaveIsSet = true;
	foreach AllActors(class'QueenDest', QD)
	{
		bQueenDestsHave = true;
		break;
	}
	logger(LOG_Debug, "QueenDestsHave", "check for QueenDests results in" @ bQueenDestsHave);
	return bQueenDestsHave;
}



function Actor SpawnActor(class<Actor> SpawnClass, ELogType LogLevel, optional Actor SpawnOwner, optional name SpawnTag, optional Vector SpawnLocation, optional Rotator SpawnRotation) {
/******************************************************************************
Tries to spawn an Actor of the given *SpawnClass* with the property
bBlockActors=false. If spawn was successful, Actor.bBlockActors is reset to
class' default value.
Either the new Actor is returned or None if spawn failed.
******************************************************************************/
local Actor result;
local bool bOldBlockActors;
local bool bOldCollideWorld;

	bOldBlockActors = SpawnClass.Default.bBlockActors;
	bOldCollideWorld = SpawnClass.Default.bCollideWorld;
	SpawnClass.Default.bBlockActors = false;
	SpawnClass.Default.bCollideWorld = false;
	result = spawn(SpawnClass, SpawnOwner, SpawnTag, SpawnLocation, SpawnRotation);
	SpawnClass.Default.bBlockActors = bOldBlockActors;
	SpawnClass.Default.bCollideWorld = bOldCollideWorld;
	if (result == None)
		Logger(LOG_Error, "SpawnActor", "could not spawn" @ SpawnClass @ "at location" @ SpawnLocation);
	else
	{
		if ( ! class'SharedCode'.static.SetCollisionEx(result, result.bCollideActors, SpawnClass.Default.bBlockActors, result.bBlockPlayers))
			Logger(LOG_Error, "SpawnActor", "SetCollisionEx faild for" @ result);
		result.bCollideWorld = bOldCollideWorld;
		result.bBlockActors = bOldBlockActors;
		Logger(LOG_Verbose, "SpawnActor", "spawned" @ result @ "at location" @ SpawnLocation);
	}
	return result;
}



function Actor SpawnAndPasteActor(actor Template, coerce string aClassName) {
/******************************************************************************
Does nearly the same as *ReplaceWith* but returns the spawned Actor.
******************************************************************************/
local Actor A;
local class<Actor> aClass;
local bool bOldCollideActors, bOldBlockActors, bOldBlockPlayers;

	aClass = class<Actor>(DynamicLoadObject(aClassName, class'Class'));
	if (aClass == None)
	{
		logger(LOG_Error, "SpawnAndPasteActor", "class '" $ aClassName $ "' could not be loaded.");
		return None;
	}
	else
	{
		// Spawn logging is done in SpawnActor()
		A = SpawnActor(aClass, LogLevel, Template.Owner, Template.tag, Template.Location, Template.Rotation);
		if (A == None) // failed: issue an error msg and exit
		{
			logger(LOG_Error, "SpawnAndPasteActor", "could not spawn" @ aClassName @ "for Template=\"" $ Template $ "\" at" @ Template.Location);
			return None;
		}
		//Template.SetPhysics(Template.Physics);
	}
	return A;
}



defaultproperties {
	bActive=true
	bQueenDestsHaveIsSet=false
	bReplaceQueenSubClassesAlso=true
	bUseLogTimeStamp=true
	LogLevel=LOG_Info
}

